"""
Deep Research Plugin Package

Implementation of Test-Time Diffusion Deep Researcher (TTD-DR) algorithm
for comprehensive research report generation.
"""

from .research_engine import DeepResearcher

__version__ = "1.0.0"
__author__ = "OptILLM Contributors"
__description__ = "TTD-DR Implementation for Deep Research"