"""LongCePO Plugin Package

Implementation of Long-Context Cerebras Planning and Optimization method.
"""

from .main import run_longcepo

__version__ = "1.0.0"
__author__ = "Cerebras"
__all__ = ['run_longcepo']